# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Encode and decode between application data and portable representation.

This module provides the basis for implementation of all codecs, i.e. the
kipjak objects that can encode and decode items of application data.

Encoding and decoding is split into a 2-layer process. The upper layer
deals with the conversion of application data to and from an internal
generic form. The lower layer is dedicated to the rendering of generic
forms into specific representations and the subsequent parsing of those
representations back into generic forms.

The :class:`~.Codec` class provides the upper layer. Lower
layers are provided by classes such as :class:`~.CodecJson`.
The latter derives from the former, inheriting 2 important methods -
:meth:`~.Codec.encode` and :meth:`~.Codec.decode`.

	.. code-block:: python

		# Define the wrapper around the JSON encoding
		# primitives.
		class CodecJson(Codec):

These 2 methods manage the combination of the two layers, presenting an
encoding-independent interface for all serialization activities within
the library.

"""

# .. autoclass:: CodecError
# .. autoclass:: CodecUsageError
# .. autoclass:: CodecRuntimeError
# .. autoclass:: Codec
#	:members:
#	:no-undoc-members:
# .. autofunction:: python_to_word
# .. autofunction:: word_to_python

__docformat__ = 'restructuredtext'

import base64
import uuid
import sys
import types
from datetime import datetime, timedelta
from copy import deepcopy
from enum import Enum


from .virtual_memory import *
from .convert_memory import *
from .convert_signature import *
from .convert_type import *
from .virtual_runtime import *
from .message_memory import *
from .make_message import *


__all__ = [
	'TypeType',
	'NoneType',
	'CodecError',
	'CodecUsageError',
	'CodecRuntimeError',
	'EnumerationRuntimeError',
	'python_to_word',
	'word_to_python',
	'Codec',

	'pass_thru',
	'w2p_string',
	'w2p_block',
	'w2p_clock',
	'w2p_span',
	'w2p_world',
	'w2p_delta',
	'w2p_uuid',
	'w2p_enumeration',
	'w2p_type',
	'w2p_portable',
	'w2p_any',

	'p2w_string',
	'p2w_block',
	'p2w_clock',
	'p2w_span',
	'p2w_world',
	'p2w_delta',
	'p2w_uuid',
	'p2w_enumeration',
	'p2w_type',
	'p2w_portable',
	'p2w_any',
]

#
#
TypeType = type
NoneType = type(None)

#
#
class CodecError(Exception):
	"""Base exception for all codec exceptions."""

	def __init__(self, note: str):
		Exception.__init__(self, note)
		self.note = note				# For quick access.

class CodecUsageError(CodecError):
	"""Cannot proceed due to its supplied environment such as unusable parameters."""

	def __init__(self, note: str, *a):
		"""Construct the exception.

		:param note: a short, helpful description
		:param a: values to be substituted into ``note``
		"""
		CodecError.__init__(self, note % a)

class CodecRuntimeError(CodecError):
	"""Failure during actual encoding or decoding, such as parsing."""

	def __init__(self, note: str, *a):
		"""Construct the exception.

		:param note: a short, helpful description
		:param a: values to be substituted into ``note``
		"""
		CodecError.__init__(self, note % a)

class EnumerationRuntimeError(Exception):
	"""Cannot encode/decode an enumeration."""

	def __init__(self, note):
		"""Construct the exception."""
		Exception.__init__(self, note)

class CircularReferenceError(Exception):
	pass

#
#
def decode_type(s):
	"""Convert the dotted string *s* to a class, or None."""
	try:
		i = s.rindex('.')
	except ValueError:
		return None
	module = s[:i]
	name = s[i + 1:]
	try:
		m = sys.modules[module]
	except KeyError:
		return None

	try:
		c = m.__dict__[name]
	except KeyError:
		return None
	return c

def encode_type(c):
	"""Convert the class *c* to a dotted string representation."""
	b = c.__art__     # kipjak runtime.
	e = '%s.%s' % (b.module, b.name)
	return e

# Transform python data to generic words. Immediately below
# are the code fragments that perform conversions from a
# specific type of python data to a declared kipjak type.
# These fragments are loaded into a jump table and the
# python_to_word function packages proper access to the
# table.
#
# All application data is reduced to an instance of the
# following types. All language and kipjak type information
# is cleared away. The generic types are;
#
# * bool .....
# * int ......
# * float ....
# * str ...... unicode
# * list ..... [v, v, v, ...]
# * dict ..... {k: v, k: v, ...}
# * none ..... null
#
# e.g.;
# * an array of 8 integers will be rendered as a list.
# * a map<string,list<int>> will be rendered as a list of pairs.
# * the dict type is reserved for rendering of structs/objects.

def pass_thru(c, p, t):
	return p

def p2w_block(c, p, t):
	w = base64.b64encode(p)
	w = w.decode(encoding='utf-8', errors='strict')
	return w

def p2w_string(c, p, t):
	w = ''
	for b in p:
		w += chr(b)
	return w

def p2w_clock(c, p, t):
	w = clock_to_text(p)
	return w

def p2w_span(c, p, t):
	w = span_to_text(p)
	return w

def p2w_world(c, p, t):
	w = world_to_text(p)
	return w

def p2w_delta(c, p, t):
	w = delta_to_text(p)
	return w

def p2w_uuid(c, p, t):
	w = uuid_to_text(p)
	return w

def p2w_enumeration(c, p, t):
	w = p.name
	return w

class NotFound(object): pass

def p2w_message(c, p, t):
	message = t.element
	rt = message.__art__
	schema = rt.schema

	w = {}
	for k, v in schema.items():
		c.walking_stack.append(k)

		# Ensure the matching pop happens.
		def get_put():
			m = getattr(p, k, NotFound)
			if m is NotFound:
				return
			elif m is None:
				if is_structural(v):
					raise ValueError(f'null structure')
			w[k] = python_to_word(c, m, v)
		get_put()
		c.walking_stack.pop()
	return w

def p2w_array(c, p, t):
	e = t.element
	n = len(p)
	s = t.size
	if n != s:
		raise ValueError(f'array [{n}] vs type [{s}]')
	w = []
	for i, y in enumerate(p):
		c.walking_stack.append(i)
		a = python_to_word(c, y, e)
		w.append(a)
		c.walking_stack.pop()
	return w

def p2w_vector(c, p, t):
	e = t.element
	w = []
	for i, y in enumerate(p):
		c.walking_stack.append(i)
		a = python_to_word(c, y, e)
		w.append(a)
		c.walking_stack.pop()
	return w

def p2w_set(c, p, t):
	e = t.element
	w = []
	for y in p:
		a = python_to_word(c, y, e)
		w.append(a)
	return w

def p2w_map(c, p, t):
	k_t = t.key
	v_t = t.value
	w = []
	for k, v in p.items():
		a = python_to_word(c, k, k_t)
		b = python_to_word(c, v, v_t)
		w.append([a, b])
	return w

def p2w_deque(c, p, t):
	e = t.element
	w = []
	for y in p:
		a = python_to_word(c, y, e)
		w.append(a)
	return w

def p2w_pointer(c, p, t):
	k = id(p)
	try:
		a = c.aliased_pointer[k]
		return a[0]
	except KeyError:
		pass

	composite = '%d:%s' % (c.pointer_alias, c.alias_space)
	a = [composite, None]
	c.pointer_alias += 1
	c.aliased_pointer[k] = a

	w = python_to_word(c, p, t.element)
	a[1] = w
	c.any_stack[-1].add(composite)
	return a[0]

def p2w_type(c, p, t):
	if isinstance(p, TypeNotBound):
		return p.module_name
	b = p.__art__
	w = b.path
	return w

def p2w_portable(c, p, t):
	if isinstance(p, Portable):
		w = portable_to_signature(p)
	elif p is None:
		w = None
	else:
		raise ValueError(f'not a portable type')
	return w

def p2w_target(c, p, t):
	# TODO
	# Perhaps the JSON encoder passes this
	# through as a list anyway. No need for
	# transform?
	w = list(p)
	return w

def unique_key(c, i):
	k = f'{c.opcode}:{i}'	# Every address is unique to a call to encode() and
	return k				# an instance of Address within that operation.

def p2w_address(c, p, t):
	if c.address_book is not None:
		# External addressing enabled. Add the address to
		# the book and return the key.
		i = len(c.address_book)
		k = unique_key(c, i)
		c.address_book[k] = p
		return k
	# Check to see if an address is being passed back over the
	# connection it arrived on. Prevents trombone behaviour.
	# Detection happens *here* at the remote end of the trombone
	# because this codec knows it is sending an address back to
	# where it came from. Add the invalid point id. See w2p_address.
	if c.return_proxy is not None:
		a = c.return_proxy
		if p[-1] == a:
			# Need to advise remote that address
			# is returning to where it came from.
			w = list(p[:-1])	# DROP THIS PROXY
			w.append(0)			# SPECIAL MARK
			return w
	w = list(p)
	return w

def p2w_any(c, p, t):
	if isinstance(p, Incognito):			# Created during a previous decoding operation.
		type_name = p.type_name				# Just need to put the same materials back on
		encoded_word = p.decoded_word		# the wire.
		if p.saved_pointers:
			c.portable_pointer.update(p.saved_pointers)				# Include upstream pointer materials.
			saved_pointers = [k for k in p.saved_pointers.keys()]	# List of pointers.
		else:
			saved_pointers = []
		if c.address_book is not None:
			c.address_book.update(p.address_book)	# Drop the decoded addresses into the outgoing book.

	elif hasattr(p, '__art__'):				# A message is outbound.
		s = c.any_stack
		s.append(set())
		u = UserDefined(type(p))
		type_name = python_to_word(c, u, Portable())
		encoded_word = python_to_word(c, p, u)
		n = s.pop()
		s[-1].update(n)
		saved_pointers = [x for x in n]

	elif isinstance(p, tuple) and len(p) == 2 and isinstance(p[1], Portable):	# An anonymous type.
		s = c.any_stack
		s.append(set())
		type_name = python_to_word(c, p[1], Portable())
		encoded_word = python_to_word(c, p[0], p[1])
		n = s.pop()
		s[-1].update(n)
		saved_pointers = [x for x in n]

	else:
		raise ValueError(f'unexpected any value {p}')

	return [type_name, encoded_word, saved_pointers]

# Map the python+portable pair to a dedicated
# transform function.
p2w = {
	# Direct mappings.
	(bool, Boolean): pass_thru,
	(int, Byte): pass_thru,
	(bytes, Character): p2w_string,
	(str, Rune): pass_thru,
	(int, Integer2): pass_thru,
	(int, Integer4): pass_thru,
	(int, Integer8): pass_thru,
	(int, Unsigned2): pass_thru,
	(int, Unsigned4): pass_thru,
	(int, Unsigned8): pass_thru,
	(float, Float4): pass_thru,
	(float, Float8): pass_thru,
	(Enum, Enumeration): p2w_enumeration,
	(bytearray, Block): p2w_block,
	(bytes, String): p2w_string,
	(str, Unicode): pass_thru,
	(float, ClockTime): p2w_clock,
	(float, TimeSpan): p2w_span,
	(datetime, WorldTime): p2w_world,
	(timedelta, TimeDelta): p2w_delta,
	(uuid.UUID, UUID): p2w_uuid,
	(list, ArrayOf): p2w_array,
	(list, VectorOf): p2w_vector,
	(set, SetOf): p2w_set,
	(dict, MapOf): p2w_map,
	(deque, DequeOf): p2w_deque,
	(TypeType, Type): p2w_type,
	(Portable, Portable): p2w_portable,
	(tuple, TargetAddress): p2w_target,
	(tuple, Address): p2w_address,

	# PointerTo - can be any of the above.
	(bool, PointerTo): p2w_pointer,
	(int, PointerTo): p2w_pointer,
	(float, PointerTo): p2w_pointer,
	(bytearray, PointerTo): p2w_pointer,
	(bytes, PointerTo): p2w_pointer,
	(str, PointerTo): p2w_pointer,
	# ClockTime and TimeDelta. Float/ptr already in table.
	# (float, PointerTo): p2w_pointer,
	# (float, PointerTo): p2w_pointer,
	(datetime, PointerTo): p2w_pointer,
	(timedelta, PointerTo): p2w_pointer,
	(uuid.UUID, PointerTo): p2w_pointer,
	(list, PointerTo): p2w_pointer,
	(set, PointerTo): p2w_pointer,
	(dict, PointerTo): p2w_pointer,
	(deque, PointerTo): p2w_pointer,
	(TypeType, PointerTo): p2w_pointer,
	(tuple, PointerTo): p2w_pointer,
	(Message, PointerTo): p2w_pointer,

	# Two mechanisms for including messages
	(Message, UserDefined): p2w_message,
	(Message, Any): p2w_any,
	(tuple, Any): p2w_any,

	# Support for Word, i.e. passthru anything
	# that could have been produced by the functions
	# above. No iterating nested layers.

	(bool, Word): pass_thru,
	(int, Word): pass_thru,
	(float, Word): pass_thru,
	# (bytearray, Word): pass_thru,
	# (bytes, Word): pass_thru,
	(str, Word): pass_thru,
	(list, Word): pass_thru,
	(dict, Word): pass_thru,
	# set, tuple - do not appear in generic

	# Provide for null values being
	# presented for different universal
	# types.

	(NoneType, Boolean): pass_thru,
	(NoneType, Byte): pass_thru,
	(NoneType, Character): pass_thru,
	(NoneType, Rune): pass_thru,
	(NoneType, Integer2): pass_thru,
	(NoneType, Integer4): pass_thru,
	(NoneType, Integer8): pass_thru,
	(NoneType, Unsigned2): pass_thru,
	(NoneType, Unsigned4): pass_thru,
	(NoneType, Unsigned8): pass_thru,
	(NoneType, Float4): pass_thru,
	(NoneType, Float8): pass_thru,
	(NoneType, Block): pass_thru,
	(NoneType, String): pass_thru,
	(NoneType, Unicode): pass_thru,
	(NoneType, ClockTime): pass_thru,
	(NoneType, TimeSpan): pass_thru,
	(NoneType, WorldTime): pass_thru,
	(NoneType, TimeDelta): pass_thru,
	(NoneType, UUID): pass_thru,
	(NoneType, Enumeration): pass_thru,
	# DO NOT ALLOW
	# (NoneType, UserDefined): pass_thru,
	# (NoneType, ArrayOf): pass_thru,
	# (NoneType, VectorOf): pass_thru,
	# (NoneType, SetOf): pass_thru,
	# (NoneType, MapOf): pass_thru,
	# (NoneType, DequeOf): pass_thru,
	(NoneType, PointerTo): pass_thru,
	(NoneType, Type): pass_thru,
	(NoneType, TargetAddress): pass_thru,
	(NoneType, Address): pass_thru,
	(NoneType, Word): pass_thru,
	(NoneType, Any): pass_thru,
}

def python_to_word(c, p, t):
	"""Generate word equivalent for the supplied application data.

	:param c: the active codec
	:type c: a kipjak Codec
	:param p: the data item
	:type p: application data
	:param t: the portable description of `p`.
	:type t: a portable expression
	:return: a generic word, ready for serialization.
	"""
	try:
		if is_message(p):
			a = Message
		elif isinstance(p, Enum):
			a = Enum
		elif isinstance(p, Portable):
			a = Portable
		else:
			a = getattr(p, '__class__')
	except AttributeError:
		a = None

	try:
		b = t.__class__		 # One of the portable types.
	except AttributeError:
		b = None

	if a is None:
		if b is None:
			raise TypeError('unusable value and type')
		raise TypeError(f'value with type "{b.__name__}" is unusable')
	elif b is None:
		raise TypeError(f'value "{a.__name__}" is unusable')

	try:
		f = p2w[a, b]
	except KeyError:
		raise TypeError(f'no transform {a.__name__}/{b.__name__}')

	# Apply the transform function
	return f(c, p, t)

# From generic data (after parsing) to final python
# representation in the application.

def w2p_string(c, w, t):
	b = bytearray()
	for c in w:
		b.append(ord(c))
	return bytes(b)

def w2p_block(c, w, t):
	p = base64.b64decode(w)
	return bytearray(p)

def w2p_clock(c, w, t):
	p = text_to_clock(w)
	return p

def w2p_span(c, w, t):
	p = text_to_span(w)
	return p

def w2p_world(c, w, t):
	p = text_to_world(w)
	return p

def w2p_delta(c, w, t):
	p = text_to_delta(w)
	return p

def w2p_uuid(c, w, t):
	p = text_to_uuid(w)	 # Throws a ValueError.
	return p

def w2p_enumeration(c, w, t):
	try:
		p = t.element[w]
	except KeyError:
		m = t.element.__name__
		raise ValueError(f'undefined enum "{m}.{w}"')
	return p

def w2p_message(c, w, t):
	u = t.element
	rt = t.element.__art__
	schema = rt.schema
	# Use the full set of names from the schema
	# to pull named values from the dict. If the
	# name is not present this is assumed to be a
	# case of skipping the encode of null values.
	# Also the scenario enforced by the version
	# slicing.
	p = u()
	for k, v in schema.items():
		c.walking_stack.append(k)

		def get_put():
			d = w.get(k, NotFound)
			if d is NotFound:		# Expected by schema, not present. Ignore.
				return
			elif d is None:
				if is_structural(v):
					raise ValueError(f'null structure')
				return

			def patch(p, k, a):
				setattr(p, k, a)
			try:
				a = word_to_python(c, d, v)
				setattr(p, k, a)
			except CircularReferenceError:
				c.patch_work.append([d, p, k, patch])
		get_put()

		c.walking_stack.pop()

	return p

def w2p_pointer(c, a, t):
	# None is handled in the table
	# (NoneType, PointerTo): pass_thru

	# 1. Is this a recursive visit to a - throw.
	# 2. Has the address word aleady been decoded.
	# 3. Find the shipped generic word.
	# 4. Guarded decode of generic word.
	# 5. Remember the decode.

	if a in c.pointer_reference:
		raise CircularReferenceError()

	try:
		p = c.decoded_pointer[a]
		return p
	except KeyError:
		pass

	try:
		w = c.portable_pointer[a]
	except KeyError:
		raise ValueError('dangling pointer')

	c.pointer_reference.add(a)
	p = word_to_python(c, w, t.element)
	c.pointer_reference.remove(a)

	c.decoded_pointer[a] = p
	return p

class TypeNotBound:
	def __init__(self, module_name: str=None):
		self.module_name = module_name

bind_message(TypeNotBound)

def w2p_type(c, w, t):
	p = decode_type(w)
	return p or TypeNotBound(module_name=w)

def w2p_portable(c, w, t):
	p = lookup_signature(w)
	if p is None:
		raise ValueError(f'unknown type "{w}"')
	return p

def w2p_array(c, w, t):
	e = t.element
	n = len(w)
	x = 0
	s = t.size
	if n > s:
		# Inbound value is longer than the target. Ignore the
		# additional items.
		# Previously there was an exception;
		# raise ValueError('array size vs specification - %d/%d' % (n, s))
		n = s
	elif n < s:
		# Inbound value is shorter than the target. Add a tail of default
		# values. This supports the guarantee that all arrays are the
		# expected size and each element is a reasonable default, i.e.
		# as defined by the element expression.
		x = s - n
		v = make(e)	 # Form the first default from the expression.
	p = []
	for i in range(n):
		d = w[i]
		c.walking_stack.append(i)

		def patch(p, i, a):
			p[i] = a
		try:
			a = word_to_python(c, d, e)
			p.append(a)
		except CircularReferenceError:
			p.append(None)
			c.patch_work.append([d, p, i, patch])
		c.walking_stack.pop()

	for i in range(x):
		p.append(v)
		v = deepcopy(v)
	return p

def w2p_vector(c, w, t):
	e = t.element
	p = []
	for i, d in enumerate(w):
		c.walking_stack.append(i)

		def patch(p, i, a):
			p[i] = a
		try:
			a = word_to_python(c, d, e)
			p.append(a)
		except CircularReferenceError:
			p.append(None)
			c.patch_work.append([d, p, i, patch])
		c.walking_stack.pop()
	return p

def w2p_set(c, w, t):
	e = t.element
	p = set()
	for d in w:
		a = word_to_python(c, d, e)
		p.add(a)
	return p

def w2p_map(c, w, t):
	k_t = t.key
	v_t = t.value
	p = {}
	for d in w:
		k = word_to_python(c, d[0], k_t)

		def patch(p, k, a):
			p[k] = a
		try:
			v = word_to_python(c, d[1], v_t)
			p[k] = v
		except CircularReferenceError:
			c.patch_work.append([d[1], p, k, patch])
	return p

def w2p_deque(c, w, t):
	e = t.element
	p = deque()
	for i, d in enumerate(w):
		def patch(p, i, a):
			p[i] = a
		try:
			a = word_to_python(c, d, e)
			p.append(a)
		except CircularReferenceError:
			p.append(None)
			c.patch_work.append([d, p, i, patch])
	return p

def w2p_target(c, w, t):
	if c.local_termination is None:
		p = tuple(w)
	elif len(w) < 2:
		p = c.local_termination,
	else:
		p = tuple(w[:-1])
	return p

def w2p_address(c, w, t):
	if c.address_book is not None:
		p = c.address_book[w]
		return p

	if c.return_proxy is not None:
		# Clean out any trombone detected
		# in the remote. See p2w_address.
		a = w[-1]
		if a == 0:	  # SPECIAL MARK
			# Address has returned home
			# No need to append a trip back
			# over this connection.
			w.pop()
			if len(w) == 0:
				# Except when its the blind address.
				w.append(c.local_termination)
		else:
			w.append(c.return_proxy)
	p = tuple(w)	# Now convert.
	return p

def w2p_null_pointer(c, w, t):
	return [0, None]

# Covert inbound 3-word tuple into the original object
def w2p_any(c, w, t):
	type_name = w[0]
	decoded_word = w[1]
	saved_pointers = w[2]

	s = lookup_signature(type_name)
	if s is None:
		y = c.portable_pointer
		h = [x for x in saved_pointers if x not in y]
		if h:
			raise ValueError(f'missing pointers')
		m = {x: y[x] for x in saved_pointers}
		p = Incognito(type_name, decoded_word, m, c.address_book)

	elif isinstance(s, UserDefined):
		p = word_to_python(c, decoded_word, s)

	else:
		p = word_to_python(c, decoded_word, s)
		p = (p, s)

	return p

#
#
w2p = {
	# Direct mappings. Left part of key is
	# the type used in a generic representation to
	# pass the intended kipjak type, i.e. if we are
	# expecting b then it should arrive as an a.
	(bool, Boolean): pass_thru,
	(int, Byte): pass_thru,
	(str, Character): w2p_string,
	(str, Rune): pass_thru,
	(int, Integer2): pass_thru,
	(int, Integer4): pass_thru,
	(int, Integer8): pass_thru,
	(int, Unsigned2): pass_thru,
	(int, Unsigned4): pass_thru,
	(int, Unsigned8): pass_thru,
	(float, Float4): pass_thru,
	(float, Float8): pass_thru,
	(str, Block): w2p_block,
	(str, String): w2p_string,
	(str, Unicode): pass_thru,
	(str, ClockTime): w2p_clock,
	(str, TimeSpan): w2p_span,
	(str, WorldTime): w2p_world,
	(str, TimeDelta): w2p_delta,
	(str, UUID): w2p_uuid,
	(str, Enumeration): w2p_enumeration,
	(list, ArrayOf): w2p_array,
	(list, VectorOf): w2p_vector,
	(list, SetOf): w2p_set,
	(list, MapOf): w2p_map,
	(list, DequeOf): w2p_deque,
	(list, TargetAddress): w2p_target,
	(list, Address): w2p_address,
	(str, Address): w2p_address,
	(str, PointerTo): w2p_pointer,

	# Two mechanisms for including messages
	# and the representation of message type.
	(dict, UserDefined): w2p_message,
	(list, Any): w2p_any,
	(str, Type): w2p_type,
	(str, Portable): w2p_portable,

	# Support for Word, i.e. passthru anything
	# that could have been produced by generic
	# layer. No iterating nested layers.

	(bool, Word): pass_thru,
	(int, Word): pass_thru,
	(float, Word): pass_thru,
	(str, Word): pass_thru,
	(list, Word): pass_thru,
	(dict, Word): pass_thru,

	# Provide for null values being
	# presented for different universal
	# types.

	(NoneType, Boolean): pass_thru,
	(NoneType, Byte): pass_thru,
	(NoneType, Character): pass_thru,
	(NoneType, Rune): pass_thru,
	(NoneType, Integer2): pass_thru,
	(NoneType, Integer4): pass_thru,
	(NoneType, Integer8): pass_thru,
	(NoneType, Unsigned2): pass_thru,
	(NoneType, Unsigned4): pass_thru,
	(NoneType, Unsigned8): pass_thru,
	(NoneType, Float4): pass_thru,
	(NoneType, Float8): pass_thru,
	(NoneType, Block): pass_thru,
	(NoneType, String): pass_thru,
	(NoneType, Unicode): pass_thru,
	(NoneType, WorldTime): pass_thru,
	(NoneType, ClockTime): pass_thru,
	(NoneType, TimeSpan): pass_thru,
	(NoneType, UUID): pass_thru,
	(NoneType, Enumeration): pass_thru,
	# DO NOT allow the automatic acceptance
	# of None as a structured value.
	# (NoneType, UserDefined): pass_thru,
	# (NoneType, ArrayOf): pass_thru,
	# (NoneType, VectorOf): pass_thru,
	# (NoneType, SetOf): pass_thru,
	# (NoneType, MapOf): pass_thru,
	# (NoneType, DequeOf): pass_thru,
	(NoneType, PointerTo): pass_thru,
	(NoneType, Type): pass_thru,
	(NoneType, Portable): pass_thru,
	(NoneType, TargetAddress): pass_thru,
	(NoneType, Address): pass_thru,
	(NoneType, Word): pass_thru,
	(NoneType, Any): pass_thru,
}

#
#
def word_to_python(c, w, t):
	"""Transform generic word to an instance of application data.

	:param c: the active codec
	:type c: a kipjak Codec
	:param w: the portable data
	:type w: generic word
	:param t: the portable description of `w`
	:type t: a portable expression
	:return: application data.
	"""
	try:
		a = w.__class__	 # The generic type.
	except AttributeError:
		a = None

	try:
		b = t.__class__	 # One of the universal types.
	except AttributeError:
		b = None

	if a is None:
		if b is None:
			raise TypeError('unusable value and type')
		raise TypeError(f'value with type "{b.__name__}" is unusable')
	elif b is None:
		raise TypeError(f'value "{a.__name__}" is unusable')

	try:
		f = w2p[a, b]
	except KeyError:
		raise TypeError(f'no transform {a.__name__}/{b.__name__}')

	return f(c, w, t)

# The base class for all codecs and essentially a
# wrapping around 2 functions;
# 1. word to text representation (w2t)
# 2. text representation to word (t2w - parsing)

STARTING_ALIAS = 1100

class Codec(object):
	"""Base class for all codecs, e.g. CodecJson."""

	def __init__(self,
			extension,
			w2t,
			t2w,
			return_proxy, local_termination, pretty_format, decorate_names):
		"""Construct the codec.

		:param extension: the additional text added to file names, e.g. ``json``
		:type extension: str
		:param w2t: the low-level conversion of application data to its text representation
		:type w2t: function
		:param t2w: the low-level parsing of text back to application data.
		:type t2w: function
		:param return_proxy: an address that the codec will use to transform deserialized addresses.
		:type return_proxy: internal
		:param local_termination: an address the codec will use to transform deserialized, “to” addresses.
		:type local_termination: internal
		:param pretty_format: generate a human-readable layout, defaults to ``True``
		:type pretty_format: bool
		:param decorate_names: auto-append a dot-extension suffix, defaults to ``True``
		:type decorate_names: bool
		"""
		self.extension = extension
		self.w2t = w2t
		self.t2w = t2w

		if return_proxy is None:
			self.return_proxy = 0
		elif not isinstance(return_proxy, tuple) or len(return_proxy) != 1:
			raise CodecUsageError('unusable return proxy')
		else:
			self.return_proxy = return_proxy[0]

		if local_termination is None:
			self.local_termination = 0
		elif not isinstance(local_termination, tuple) or len(local_termination) != 1:
			raise CodecUsageError('unusable local termination')
		else:
			self.local_termination = local_termination[0]

		self.pretty_format = pretty_format
		self.decorate_names = decorate_names

		self.address_book = None

		# Encode/decode collections
		self.walking_stack = []
		self.aliased_pointer = {}		# Encoding.
		self.portable_pointer = {}		# Both.
		self.pointer_reference = set()
		self.decoded_pointer = {}		# Decoding.
		self.patch_work = []
		self.pointer_alias = STARTING_ALIAS
		self.alias_space = 'default'

	def encode(self, value, expression, address_book=None):
		"""Encode an application value to its portable representation.

		:param value: a runtime application value
		:type value: a type consistent with the specified `expression`
		:param expression: a formal description of the `value`
		:type expression: :ref:`type expression<type-reference>`
		:param version: an explicit version override
		:type version: string
		:return: a portable representation of the `value`
		:rtype: str
		"""
		self.address_book = address_book
		self.walking_stack = []				# Breadcrumbs for m.a[0].f.c[1] tracking.
		self.aliased_pointer = {}			# Pointers encountered in value.
		self.portable_pointer = {}			# Pointers accumulated from Incognitos.
		self.any_stack = [set()]
		self.pointer_alias = STARTING_ALIAS

		u4 = uuid.uuid4()
		self.alias_space = str(u4)
		self.opcode = str(u4)

		try:
			# Convert the value to a generic intermediate
			# representation.

			w = python_to_word(self, value, expression)
		except (AttributeError, TypeError, ValueError, IndexError, KeyError, ConversionEncodeError) as e:
			s = str(e)
			nesting = self.nesting()
			if len(nesting) == 0:
				raise CodecRuntimeError(f'cannot encode ({s})')
			raise CodecRuntimeError(f'cannot encode, near "{nesting}" ({s})')

		# Create a dict with value, address and version.
		shipment = {'value': w}
		if len(self.aliased_pointer) > 0:
			# New pointers in the p2w transformations. Need to add them
			# to the older accumulated pointer materials (i.e. Incognitos).
			a = {v[0]: v[1] for _, v in self.aliased_pointer.items()}
			self.portable_pointer.update(a)

		if len(self.portable_pointer) > 0:
			# Pointers in the outbound encoding. Need to
			# flatten then into generic form.
			shipment['pointer'] = [[k, v] for k, v in self.portable_pointer.items()]

		try:
			# Convert generic form to portable
			# representation.

			s = self.w2t(self, shipment)
		except (AttributeError, TypeError, ValueError, IndexError, KeyError) as e:
			e = str(e)
			raise CodecRuntimeError(f'cannot encode ({e})')
		return s

	def decode(self, representation, expression, address_book=None):
		"""Decode a representation to its final application form.

		:param representation: the result of a previous encode operation
		:type representation: str
		:param expression: a formal description of portable
		:type expression: a :ref:`type expression<type-reference>`
		:return: an application value
		"""
		self.address_book = address_book

		self.walking_stack = []		 # Breadcrumbs for m.a[0].f.c[1] tracking.
		self.portable_pointer = {}	  # Shipped pointer materials.
		self.decoded_pointer = {}	   # Pointers transformed to final type.
		self.patch_work = []

		try:
			# Convert portable representation into generic
			# intermediate form.
			shipment = self.t2w(self, representation)
		except (AttributeError, TypeError, ValueError, IndexError, KeyError) as e:
			s = str(e)
			raise CodecRuntimeError(f'cannot decode ({s})')

		# Now everything is in the generic form. Need to rebuild python
		# types in steps due to work around pointers.
		def decode(w, expression):
			self.walking_stack = []		 # Error tracking.
			self.pointer_reference = set()
			try:
				p = word_to_python(self, w, expression)
			except (AttributeError, TypeError, ValueError, IndexError, KeyError, ConversionDecodeError) as e:
				s = str(e)
				text = self.nesting()
				if len(text) == 0:
					raise CodecRuntimeError(f'cannot decode ({s})')
				raise CodecRuntimeError(f'cannot decode, near "{text}" ({s})')
			return p

		try:
			# Pull the address pointer-to materials out
			# and save into convenient map. Does not transform
			# into final types.
			flat = shipment['pointer']
			portable = decode(flat, MapOf(Word(), Word()))
			self.portable_pointer.update(portable)  # Use EXISTING portable_pointer
		except KeyError:
			pass
		except (AttributeError, TypeError, ValueError, IndexError):
			raise CodecRuntimeError('cannot decode (not the output of an encoding?)')

		try:
			w = shipment['value']
		except KeyError:
			raise CodecRuntimeError('cannot decode (no "value" available)')

		# Decode the word to its final application resting-place. This performs
		# transforms into final types, including the pointer materials. Backpatch
		# any circular references.
		p = decode(w, expression)
		for b in self.patch_work:
			decoded = self.decoded_pointer[b[0]]
			r, k = b[1], b[2]
			f = b[3]
			f(r, k, decoded)

		return p

	def nesting(self):
		"""Use the internal stack to generate a data path."""
		p = ''
		for s in self.walking_stack:
			if isinstance(s, int):
				p += '[%d]' % (s,)
			elif isinstance(s, str):
				if len(p) > 0:
					p += '.'
				p += '%s' % (s,)
			else:
				p += '<?>'
		return p

	def full_name(self, name):
		"""Augment the name with an extension, as appropriate."""
		if not self.decorate_names:
			return name
		if name[-1] == '.':
			return name[:-1]
		s = '%s.%s' % (name, self.extension)
		return s



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Concurrency and support of same.

Objects and classes that allow for the construction of
complex distributed operations.
"""
__docformat__ = 'restructuredtext'

from .message_memory import *
from .convert_type import *
from .virtual_memory import *
from .virtual_point import *
from .virtual_runtime import *
from .point_runtime import *
from .point_machine import *
from .bind_type import *

__all__ = [
	'CreateFrame',
	'GetResponse',
	'Delay',
	'Concurrently',
	'Sequentially',
]

#
class CreateFrame(object):
	"""Capture values needed for async object creation.

	:param object_type: type to be created
	:type object_type: :ref:`object type<kj-object-type>`
	:param args: positional args to be passed on creation
	:param kw: named values to be passed on creation
	"""
	def __init__(self, object_type, *args, **kw):
		self.object_type = object_type
		self.args = args
		self.kw = kw

#
class GetResponse(Point, Stateless):
	"""
	Delegated request-response.

	Send the request to the given address and expect a
	response. Pass the response back as the return value
	for this object.

	:param request: message to be sent
	:type request: :ref:`message<kj-message>`
	:param server_address: where to send the message
	:param seconds: acceptable delay
	"""
	def __init__(self, request, server_address: Address, seconds: float=None):
		Point.__init__(self)
		Stateless.__init__(self)
		self.request = request
		self.server_address = server_address
		self.seconds = seconds

def GetResponse_Start(self, message):
	self.send(self.request, self.server_address)	# Request.
	if self.seconds is not None:
		self.start(T1, self.seconds)

def GetResponse_T1(self, message):						# Too slow.
	self.complete(TimedOut(message))

def GetResponse_Stop(self, message):					# Interruption.
	self.complete(Aborted())

def GetResponse_Unknown(self, message):	# Assumed to be response. Use as completion.
	m = cast_to(message, self.received_type)
	self.complete(m)

GET_RESPONSE_DISPATCH = [
	Start,
	T1,
	Stop,
	Unknown,
]

bind_stateless(GetResponse, GET_RESPONSE_DISPATCH, thread='get-response', return_type=Any())

#
class Delay(Point, Stateless):
	"""Object that does nothing for the specified number of seconds."""
	def __init__(self, seconds=None):
		Point.__init__(self)
		Stateless.__init__(self)
		self.seconds = seconds

def Delay_Start(self, message):
	self.start(T1, self.seconds)

def Delay_T1(self, message):						# Too slow.
	self.complete(TimedOut(message))

def Delay_Stop(self, message):					# Interruption.
	self.complete(Aborted())

DELAY_DISPATCH = [
	Start,
	T1,
	Stop,
]

bind_stateless(Delay, DELAY_DISPATCH, thread='delay')

#
class Concurrently(Point, Stateless):
	"""
	Delegated, multi-way request-response.

	Manage one or more concurrent requests. Terminate on completion of full set, or	a timer.
	Accepts a mixed tuple, i.e. either request-address pairs or :class:`~.CreateFrame` objects.

	:param get: a list of request-address pairs or frames
	:type get: tuple
	:param seconds: acceptable delay
	"""
	def __init__(self, *get, seconds: float=None):
		Point.__init__(self)
		Stateless.__init__(self)
		self.get = get		# List of object descriptions.
		self.count = len(get)		# Save for countdown.
		self.seconds = seconds
		self.orderly = [None] * self.count	# Prepare the completion list.

def Concurrently_Start(self, message):
	if self.count < 1:
		self.complete(self.orderly)		# Nothing to do.

	def collate(self, response, kv):			# Place the completion in its proper slot.
		i = kv.i
		self.orderly[i] = cast_to(response, self.returned_type)
		self.count -= 1
		if self.count < 1:
			self.complete(self.orderly)

	# Create an object for each slot. Allow full object spec
	# or the request-address tuple.
	for i, p in enumerate(self.get):
		if isinstance(p, CreateFrame):
			a = self.create(p.object_type, *p.args, **p.kw)
		elif isinstance(p, tuple) and len(p) == 2:
			r, s = p
			a = self.create(GetResponse, r, s)		# Provide the object for simple request-response exchange.
		else:
			self.complete(Faulted(f'unexpected frame/request [{i}]'))

		self.on_return(a, collate, i=i)

	if self.seconds is not None:
		self.start(T1, self.seconds)

def Concurrently_T1(self, message):
	self.abort(TimedOut(message))

def Concurrently_Stop(self, message):
	self.abort(Aborted())

def Concurrently_Returned(self, message):
	d = self.debrief()
	if isinstance(d, OnReturned):
		d(self, message)
		return

	self.complete(Faulted(f'unexpected get completion'))

CONCURRENTLY_DISPATCH = [
	Start,
	Returned,
	T1,
	Stop,
]

bind_stateless(Concurrently, CONCURRENTLY_DISPATCH, thread='concurrently', return_type=VectorOf(Any()))


#
class Sequentially(Point, Stateless):
	"""Object that iterates multiple objects and produces list of completions."""
	def __init__(self, *get, seconds=None):
		Point.__init__(self)
		Stateless.__init__(self)
		self.get = get
		self.pointer = iter(self.get)
		self.seconds = seconds
		self.orderly = []

	def next_step(self):
		try:
			p = next(self.pointer)
		except StopIteration:
			self.complete(self.orderly)

		if isinstance(p, CreateFrame):
			a = self.create(p.object_type, *p.args, **p.kw)
		elif isinstance(p, tuple) and len(p) == 2:
			r, s = p
			a = self.create(GetResponse, r, s)		# Provide the object for simple request-response exchange.
		else:
			i = len(self.orderly)
			self.complete(Faulted(f'unexpected sequence item [{i}]'))
		return a

def Sequentially_Start(self, message):
	def step(self, value, kv):
		if isinstance(value, Faulted):
			self.complete(value)
		self.orderly.append(value)

		a = self.next_step()
		self.on_return(a, step)

	if self.seconds is not None:
		self.start(T1, self.seconds)

	a = self.next_step()
	self.on_return(a, step)

def Sequentially_T1(self, message):
	self.abort(TimedOut(message))

def Sequentially_Stop(self, message):
	self.abort(Aborted())

def Sequentially_Returned(self, message):
	d = self.debrief()
	if isinstance(d, OnReturned):
		d(self, message)
		return

	self.complete(Faulted(f'unexpected sequence completion'))

SEQUENTIALLY_DISPATCH = [
	Start,
	T1,
	Stop,
	Returned,
]

bind_stateless(Sequentially, SEQUENTIALLY_DISPATCH, thread='sequentially')



# test_library_machine.py
import kipjak as kj

from test_api import *
from test_function import *

# A bare-bones, machine implementation
# of test_library.py.
class Library(kj.Threaded, kj.Stateless):
	def __init__(self):
		kj.Threaded.__init__(self)
		kj.Stateless.__init__(self)

def Library_Start(self, message):
	pass

def Library_Xy(self, message):
	table = texture(self, x=message.x, y=message.y)
	self.send(kj.cast_to(table, table_type), self.return_address)

def Library_Stop(self, message):
	self.complete(kj.Aborted())

kj.bind(Library, (kj.Start, Xy, kj.Stop), entry_point=[Xy,])

if __name__ == '__main__':
	kj.create(Library)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Storage and recovery of application data using system files.

Two primitives - write_to_file and read_from_file - and the File
class to package the common usage patterns.

.. autofunction:: write_to_file
.. autofunction:: read_from_file

.. autoclass:: File
   :members: store, recover
   :member-order: bysource
"""

__docformat__ = 'restructuredtext'

__all__ = [
	'File',
	'read_from_file',
	'write_to_file',
]

import os

from .virtual_memory import *
from .message_memory import *
from .convert_type import *
from .make_message import *
from .virtual_codec import *
from .json_codec import *

#
#
class File(object):
	"""Store and recover application values using files.

	:param name: name of the file
	:param tip: type expression for the content
	:type tip: :ref:`tip<type-reference>`
	:param encoding: selection of representation, defaults to ``CodecJson``
	:type encoding: class
	:param create_default: return default instance if file not found on read, defaults to ``False``
	:param pretty_format: generate human-readable file contents, defaults to ``True``
	:param decorate_names: auto-append an encoding-dependent extension to the file name, defaults to ``True``
	"""
	def __init__(self, name: str, tip, encoding=None, create_default: bool=False, pretty_format: bool=True, decorate_names: bool=True):
		"""Not published."""
		self.name = name

		if tip is None:
			self.file_type = None
		elif isinstance(tip, Portable):
			self.file_type = tip
		elif hasattr(tip, '__art__'):
			self.file_type = UserDefined(tip)
		else:
			self.file_type = lookup_type(tip)

		self.encoding = encoding
		self.create_default = create_default
		self.pretty_format = pretty_format
		self.decorate_names = decorate_names

	def store(self, value, as_name=None, as_path=None):
		"""Generate a representation of ``value`` and write to the saved ``name``.

		:param value: any application value
		:type value: matching the :ref:`tip<type-reference>` saved on construction
		"""
		if as_path:
			if as_name:
				name = os.path.join(as_path, as_name)
			else:
				h, t = os.path.split(self.name)
				name = os.path.join(as_path, t)
		elif as_name:
			h, t = os.path.split(self.name)
			name = os.path.join(h, as_name)
		else:
			name = self.name

		write_to_file(value, name, self.file_type, encoding=self.encoding,
			decorate_names=self.decorate_names, pretty_format=self.pretty_format)

	def recover(self):
		"""Read from the saved ``name``, parse and marshal into an application value.

		:return: any application value
		:rtype: matching the saved :ref:`tip<type-reference>`
		"""
		try:
			r = read_from_file(self.file_type, self.name, encoding=self.encoding, decorate_names=self.decorate_names)
		except FileNotFoundError:
			if self.create_default:
				c = self.file_type
				a = make(c)
				return a
			raise
		return r

	def full_name(self, encoding=None):
		encoding = encoding or self.encoding or CodecJson
		encoding = encoding(decorate_names=self.decorate_names)
		return encoding.full_name(self.name)

	def remove(self):
		name = self.full_name()
		try:
			os.remove(name)
		except FileNotFoundError:
			return False
		return True

# The primitives.
def read_from_file(expression, name, encoding=None, **kv):
	"""Recover an application object from the representation loaded from the named file."""
	encoding = encoding or CodecJson
	encoding = encoding(**kv)

	# Add the encoding suffix according
	# to automation settings.
	name = encoding.full_name(name)

	with open(name, 'r') as f:
		s = f.read()

	d = encoding.decode(s, expression)
	return d

#
#
def write_to_file(a, name, expression=None, encoding=None, pretty_format=True, **kv):
	"""Write a representation of the application date into the named file."""
	kv['pretty_format'] = pretty_format
	encoding = encoding or CodecJson
	encoding = encoding(**kv)

	# Add the encoding suffix according
	# to automation settings.
	name = encoding.full_name(name)

	if expression is None:
		if not is_message(a):
			raise CodecUsageError(f'encoding of unregistered message to file "{name}"')
		expression = UserDefined(a.__class__)

	s = encoding.encode(a, expression)

	with open(name, 'w') as f:
		f.write(s)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementation of the JSON codec.

The plug-in part of the encoding/decoding stack and the default. Involves
two functions that transform between the internal, flattened memory
representation and the actual encoding. With a little luck there is not
much more to do than call the read/write, or load/save methods on the
encoding-specific parser module.

* word_to_json - generate the JSON representation of an application word.
* json_to_word - recovers an application word from a JSON representation.

Also refer to;

* https://tools.ietf.org/html/rfc7159
* https://www.json.org/json-en.html
"""

__docformat__ = 'restructuredtext'

# .. autoclass:: CodecJson
# .. autofunction:: word_to_json
# .. autofunction:: json_to_word

import json

from .virtual_memory import *
from .convert_memory import *
from .virtual_runtime import *
from .message_memory import *
from .virtual_codec import *


__all__ = [
	'word_to_json',
	'json_to_word',
	'CodecJson'
]

# Standard JSON library version of generation and
# parsing of standard-conforming text.
def word_to_json(c, w):
	"""Generate the JSON representation of a generic word.

	If the codec `pretty_format` property is true, this
	function will produce a more human-readable rendering
	of JSON.

	:param c: an active codec
	:type c: a Codec-based object
	:param w: a generic word
	:rtype: the JSON text.
	"""
	if c.pretty_format:
		j = json.dumps(w, sort_keys=True, indent=4, separators=(',', ': '))
	else:
		j = json.dumps(w, separators=(',', ':'))
	return j

# Decoding - from parsing of JSON to transformation
# into app data items.

def json_to_word(c, j):
	"""Produce a generic word from the parsing of a text JSON representation.

	:param c: an active codec
	:type c: a Codec-based instance
	:param j: the JSON text
	:type j: string
	:rtype: a generic word.
	"""
	j = json.loads(j)
	return j

#
#
class CodecJson(Codec):
	"""Encoding and decoding of JSON representations.

	This class is the default value for the ``encoding`` parameter, related
	to the various store and recover operations of the library. Refer to the
	:ref:`base codec class<codec-base>` for the significant methods.
	"""

	EXTENSION = 'json'

	def __init__(self, return_proxy=None, local_termination=None, pretty_format=False, decorate_names=True):
		"""Construct a JSON codec."""
		Codec.__init__(self,
			CodecJson.EXTENSION,
			word_to_json,
			json_to_word,
			return_proxy, local_termination, pretty_format, decorate_names)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A virtual memory model.

.. autoclass:: Boolean

.. autoclass:: Integer2
.. autoclass:: Integer4
.. autoclass:: Integer8
.. autoclass:: Unsigned2
.. autoclass:: Unsigned4
.. autoclass:: Unsigned8

.. autoclass:: Float4
.. autoclass:: Float8

.. autoclass:: Byte
.. autoclass:: Character
.. autoclass:: Rune
.. autoclass:: Block
.. autoclass:: String
.. autoclass:: Unicode

.. autoclass:: Enumeration

.. autoclass:: ClockTime
.. autoclass:: TimeSpan
.. autoclass:: WorldTime
.. autoclass:: TimeDelta

.. autoclass:: UUID

.. autoclass:: ArrayOf
.. autoclass:: VectorOf
.. autoclass:: SetOf
.. autoclass:: MapOf
.. autoclass:: DequeOf

.. autoclass:: UserDefined
.. autoclass:: Type
.. autoclass:: PointerTo

.. autoclass:: TargetAddress
.. autoclass:: Address

.. autoclass:: Word
.. autoclass:: Any

.. autofunction:: is_portable
.. autofunction:: is_container
.. autofunction:: is_portable_class
.. autofunction:: is_container_class
.. autofunction:: is_address
"""

__docformat__ = 'restructuredtext'

from collections import deque

__all__ = [
	# Classification.
	'Portable', 'Container',

	# Logical.
	'Boolean',

	# Integers.
	'Integer2', 'Integer4', 'Integer8',
	'Unsigned2', 'Unsigned4', 'Unsigned8',

	# Floats.
	'Float4', 'Float8',

	# Bytes, ASCII and unicode.
	'Byte', 'Character', 'Rune',
	'Block', 'String', 'Unicode',

	# Name for a number.
	'Enumeration',

	# Time.
	'ClockTime', 'TimeSpan',
	'WorldTime', 'TimeDelta',

	'UUID',

	# Containers.
	'ArrayOf', 'VectorOf', 'DequeOf',
	'SetOf', 'MapOf',

	'UserDefined',
	'PointerTo',
	'Any',

	# Networking.
	'TargetAddress',
	'Address',

	# Internal.
	'Type',
	'Word',

	# Supporting.
	'VIRTUAL_MEMORY_LIST',
	'VIRTUAL_MEMORY_SET',

	'deque',

	'is_portable',
	'is_container',
	'is_structural',
	'is_portable_class',
	'is_container_class',

	'NO_SUCH_ADDRESS',

	'is_address',
	'address_on_proxy',
]

class Portable(object):
	"""Base for all portables.

	This is the type that is combined with :class:`~.Word` to
	implement the :class:`~.Any` concept.
	"""

class Container(Portable):
	"""The subset of portables that hold zero or more portables."""

# A unit of memory.
class Boolean(Portable):
	"""True or false."""

class Integer2(Portable):
	"""A 2-byte, signed integer."""

class Integer4(Portable):
	"""A 4-byte, signed integer."""

class Integer8(Portable):
	"""An 8-byte, signed integer."""

class Unsigned2(Portable):
	"""A 2-byte, unsigned integer."""

class Unsigned4(Portable):
	"""A 4-byte, unsigned integer."""

class Unsigned8(Portable):
	"""An 8-byte, unsigned integer."""

class Float4(Portable):
	"""A 4-byte, floating point number."""

class Float8(Portable):
	"""An 8-byte, floating point number."""

class Byte(Portable):
	"""The smallest unit of data - 8 bit, unsigned integer."""

class Character(Portable):
	"""A byte that more often than not, contains a printable ASCII character."""

class Rune(Portable):
	"""A Unicode codepoint."""

class Block(Portable):
	"""A sequence of Byte."""

class String(Portable):
	"""A sequence of Character."""

class Unicode(Portable):
	"""A sequence of Rune."""

class Enumeration(Portable):
	"""
	Names for integers.
	"""

	def __init__(self, element):
		self.element = element

class ClockTime(Portable):
	"""The time on the wall clock as an epoch value; a float."""

class TimeSpan(Portable):
	"""Difference between two ``ClockTime`` values; a float."""

class WorldTime(Portable):
	"""The time at the Greenwich meridian; a datetime object."""

class TimeDelta(Portable):
	"""Difference between two ``WorldTime`` values; a timedelta object."""

class UUID(Portable):
	"""An RFC 4122 UUID (random); a ``uuid.UUID`` object."""

class ArrayOf(Container):
	"""A fixed-length sequence of elements.

	:param element: type of the content.
	:param size: fixed size.
	"""

	def __init__(self, element: Portable, size: int):
		"""Refer to class."""
		self.element = element
		self.size = size

class VectorOf(Container):
	"""A variable-length sequence of elements.

	:param element: type of the content.
	"""

	def __init__(self, element: Portable):
		"""Refer to class."""
		self.element = element

class DequeOf(Container):
	"""A double-ended sequence of elements.

	:param element: type of the content.
	"""

	def __init__(self, element: Portable):
		"""Refer to class."""
		self.element = element

class SetOf(Container):
	"""A collection of unique elements.

	:param element: type of the content, a hash-able value.
	"""

	def __init__(self, element: Portable):
		"""Refer to class."""
		self.element = element

class MapOf(Container):
	"""A map of unique, key-value pairs.

	:param key: type of the key, a hash-able value.
	:param value: type of the content.
	"""

	def __init__(self, key: Portable, value: Portable):
		"""Refer to class."""
		self.key = key
		self.value = value

class UserDefined(Container):
	"""A function, message class or enum class.

	:param element: Python reference
	:type element: :ref:`object type<kj-object-type>`
	"""

	def __init__(self, element):
		"""Refer to class."""
		self.element = element

class PointerTo(Container):
	"""An instance of an object that may be appear in multiple places.

	These are instances of data that are tracked by their Python id.
	When they are encoded for the purposes of networking or saving
	as a file, multiple appearances of the same id are collapsed into
	one actual data item and multiple references. This process is
	reversed during decoding, allowing the passing of graphs (linked
	lists, trees and networks) across network connections. Circular
	references are properly handled, pointer-to-pointer is not.
	"""

	def __init__(self, element: Portable):
		"""Refer to class."""
		self.element = element

class Any(Portable):
	"""
	The combination of a :class:`~.Portable` and a :class:`~.Word`.

	A portable representation of any registered type (see :func:`~.bind`
	and :func:`~.def_type`). Suitable for passing across a network
	connection or storing in a file.

	Further information can be found :ref:`here<kj-message>`.
	"""

class TargetAddress(Portable):
	"""The address of a receiving object."""

class Address(Portable):
	"""
	A runtime-generated, **kipjak** address.

	The unique identity of an asynchronous object, usually obtained from
	a call to :meth:`~.Point.create`. These are the values that can be
	passed as a destination to functions such as :meth:`~.Point.send`.

	Further information can be found :ref:`here<kj-address>`.
	"""

class Type(Portable):
	"""
	The unique, portable identity of a registered class or function.

	Transfer a function or class across a network connection. The matching
	name must exist in the receiver or a :class:`~.TypeNotBound` is created.
	"""

class Word(Portable):
	"""
	An intermediate form of application data.

	This is a transformation of application into a generic
	form before it is presented for encoding and it is the
	generic form that results from a decoding, e.g. a
	``dict`` has no representation in JSON and is converted
	into a list of pairs.
	"""

# List of the library types.
VIRTUAL_MEMORY_LIST = [
	Boolean,
	Integer2, Integer4, Integer8,
	Unsigned2, Unsigned4, Unsigned8,
	Float4, Float8,
	Byte, Character, Rune,
	String, Unicode, Block,
	Enumeration,
	ClockTime, TimeSpan,
	WorldTime, TimeDelta,
	UUID,
	UserDefined,
	ArrayOf, VectorOf, DequeOf,
	SetOf, MapOf,
	PointerTo,
	Any,
	TargetAddress, Address,
	Type,
	Word,
]

# Set of the library types.
VIRTUAL_MEMORY_SET = set(VIRTUAL_MEMORY_LIST)

# Few handy type predicates.
#
def is_portable(a):
	"""Is object *a* an instance of one of the portable types."""
	return isinstance(a, Portable)

def is_container(a):
	"""Is object *a* an instance of one of the portable container types."""
	return isinstance(a, Container)

def is_structural(a):
	"""Is object *a* an instance of one of the portable container types and not a pointer."""
	b = isinstance(a, Container) and not isinstance(a, PointerTo)
	return b

def is_portable_class(c):
	"""Is object *c* one of the portable types."""
	try:
		return issubclass(c, Portable)
	except TypeError:
		return False

def is_container_class(c):
	"""Is object *c* one of the portable container types."""
	try:
		return issubclass(c, Container)
	except TypeError:
		return False

# This is the official null address and where required the
# default value for an address.
NO_SUCH_ADDRESS = (0,)

def is_address(a):
	"""Is object *a* is a valid point address."""
	try:
		return isinstance(a, tuple) and len(a) > 0
	except (TypeError, ValueError):
		return False

def address_on_proxy(a, p):
	"""Check that address *a* refers to an object behind the proxy address, p."""
	if a[-1] == p[-1]:
		if len(p) == 1 and len(a) > 1:
			return True
	return False



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import string
import re
from urllib.parse import quote_plus, unquote_plus
from enum import Enum

from .virtual_memory import *
from .convert_memory import *
from .convert_signature import *
from .convert_type import *
from .message_memory import *
from .bind_type import *
from .command_startup import *
from .virtual_codec import *
from .json_codec import *
from .noop_codec import *
from .virtual_point import *
from .point_runtime import *
from .point_machine import *

__all__ = [
	'HttpRequest',
	'HttpResponse',
	'FormRequest',
	'HttpMethod',
	'decode_resource',
	'decode_body',
	'cannot_be_dispatched',
	'cannot_be_converted',
	'ResourceDispatch',
	'ApiServerStream',
	'ApiClientStream',
	'ApiClientSession',
]

# Inbound to API server
# or outbound from client.
class HttpRequest(object):
	"""The message generated by an HTTP client and received by an HTTP server.

	:param method: one of GET, POST, PUT, ...
	:param request_uri: the path section of a URI
	:param http: version of HTTP
	:param header: name-value pairs
	:param body: payload, value
	"""
	def __init__(self, method: str=None, request_uri: str=None, http: str=None, header: dict[str,str]=None, body: bytearray=None):
		self.method = method or 'POST'
		self.request_uri = request_uri or '/'
		self.http = http or 'HTTP/1.1'
		self.header = header or {}
		self.body = body

# Outbound from API server
# or inbound to client.
class HttpResponse(object):
	"""The message generated by an HTTP server and received by an HTTP client.

	:param http: version of HTTP
	:param status_code: 100, 200, 300, ...
	:param reason_phrase: short description
	:param header: name-value pairs
	:param body: payload, value
	"""
	def __init__(self, http: str=None, status_code: int=200, reason_phrase: str=None,
			header: dict[str,str]=None, body: bytearray=None):
		self.http = http or 'HTTP/1.1'
		self.status_code = status_code
		self.reason_phrase = reason_phrase or 'OK'
		self.header = header or {}
		self.body = body

bind(HttpRequest)
bind(HttpResponse)

# Inbound to form server.
class FormRequest(object):
	"""The message generated by an HTTP client and received by an HTTP server.

	Passing an :class:`~.ReForm` to :func:`~.listen` causes special processing
	of the request URI. The URI is deconstructed into a FormRequest and presented
	to the calling object, e.g. a server. This should be combined with use of
	the :class:`~.ResourceDispatch`, to route inbound FormRequests to a resource-
	based function.

	:param method: one of GET, POST, PUT, ...
	:param form_entry: the path section of a URI
	:param http: version of HTTP
	:param header: name-value pairs
	:param body: payload, value
	"""
	def __init__(self, method: str=None, request: str=None, form_entry: dict[str,str]=None, http: str=None, header: dict[str,str]=None, body: bytearray=None):
		self.method = method or 'POST'
		self.request = request
		self.form_entry = form_entry or {}
		self.http = http or 'HTTP/1.1'
		self.header = header or {}
		self.body = body

bind(FormRequest)

#
#
class HttpMethod(Enum):
	CONNECT=1
	DELETE=2
	GET=3
	HEAD=4
	OPTIONS=5
	PATCH=6
	POST=7
	PUT=8
	TRACE=9

def decode_resource(body, portable):
	if body is None:
		return None

	codec = CodecJson()
	json = body.decode('utf-8')
	try:
		message = codec.decode(json, portable)
	except (CodecError, OverflowError, ValueError) as e:
		s = str(e)
		raise ValueError(s)
	return message

def decode_body(portable):
	if hasattr(portable, '__art__'):
		portable = UserDefined(portable)

	def decode(body, ignored):
		if body is None:
			return None
		codec = CodecJson()
		json = body.decode('utf-8')
		try:
			message = codec.decode(json, portable)
		except (CodecError, OverflowError, ValueError) as e:
			s = str(e)
			raise ValueError(s)
		return message
	return decode

def cannot_be_dispatched(self, request=None, error=None):
	cannot = f'Cannot map request [{request.method}] "{request.request}" ({error})'
	self.warning(cannot, **request.form_entry)
	return HttpResponse(status_code=400, reason_phrase='Bad Request', body=cannot)

def cannot_be_converted(self, request=None, error=None):
	c = ','.join([f'{k}={v}' for k, v in request.form_entry.items()])
	cannot = f'Cannot complete conversion "{c}" ({error})'
	self.warning(cannot)
	return HttpResponse(status_code=400, reason_phrase='Bad Request', body=cannot)

class ResourceDispatch(object):
	'''Build up the info needed to dispatch from FormRequest to function.

	Define a global instance passing the ReForm and list of resource types.	Accept
	mappings of request-to-function, using the :meth:`~.ResourceDispatch.add` method.
	Then expect calls to :meth:`~.ResourceDispatch.lookup` that matches the runtime FormRequest to a saved
	function.
	'''
	def __init__(self, form: ReForm, resource: list[Type]):
		entry_names = form.entry_names()
		if 'resource' not in entry_names:
			raise ValueError('form has no "resource" entry')
		entry_names.remove('resource')
		self.entry_names = entry_names
		self.resource_class = {r.__name__.lower(): r for r in resource}
		self.call_frame = {}

	def add(self, resource: Type, method: HttpMethod, f, **kv):
		'''Add a mapping for a resource+method to a function.

		Compose a key based on resource, method and the set of
		expected args (``kv``). Then ``table[key] = f``. Each ``kv``
		entry is a name and a conversion function.
		'''
		expected = tuple(s in kv for s in self.entry_names)
		key = (resource, method, *expected)
		self.call_frame[key] = (f, kv)

	def lookup(self, request: FormRequest):
		'''Resolve the function to call based on info in FormRequest.

		Compose a key based on resource, method and the set of
		detected args (``request.form_entry``), convert the args from
		matched text to application values, Return function and args.
		'''
		try:
			resource = self.resource_class[request.form_entry['resource']]
			method = HttpMethod[request.method]
			expected = tuple(request.form_entry[s] is not None for s in self.entry_names)
			key = (resource, method, *expected)
			call = self.call_frame.get(key, None)
			if call is None:
				f = cannot_be_dispatched
				a = {
					'request': request,
					'error': 'no matching call signature',
				}
				return f, a
		except (ValueError, IndexError, KeyError) as e:
			s = str(e)
			f = cannot_be_dispatched
			a = {
				'request': request,
				'error': 'unusable resource, method or URI',
			}
			return f, a

		try:
			f, args = call
			a = {}
			for k, c in args.items():
				if k == 'header':
					value = request.header if c is None else c(request.header)
				elif k == 'http':
					value = request.http if c is None else c(request.http)
				elif k == 'body':
					value = request.body if c is None else c(request.body, UserDefined(resource))
				else:
					entry = request.form_entry[k]
					value = entry if c is None else c(entry)
				a[k] = value
		except (ValueError, IndexError, KeyError, TypeError) as e:
			s = str(e)
			f = cannot_be_converted
			a = {
				'request': request,
				'error': s,
			}
		return f, a

	def __call__(self, layer, request: FormRequest):
		'''Resolve the request to a function and call it.

		Make the call to lookup() and execute the resulting call materials.
		If the result of that call is not a None - send it on to the
		original client.
		'''
		return_address = layer.return_address
		resource_function, entry_args = self.lookup(request)
		response = resource_function(layer, **entry_args)
		if response is not None:
			layer.send(response, return_address)

# Conversion of messages to on-the-wire blocks, and back again.
# Sync HTTP request-response, fully typed (HTTP body).
# Accepted patterns.
REQUEST_LINE = -1
EMPTY_LINE = -2
HEADER_LINE = -3
BODY_BLOCK = -4
DISCARD_LINE = -5
RESPONSE_LINE = -6

# Character sets and constants for FSM.
HTTP_TEXT = set(ord(c) for c in string.digits) | set(ord(c) for c in string.ascii_letters) | set(ord(c) for c in string.punctuation)
HTTP_SPACE = set(ord(c) for c in [' ', '\t'])
HTTP_DIGIT = set(ord(c) for c in string.digits)
HTTP_PRINTING = HTTP_TEXT | HTTP_SPACE

CR = 13		# Carriage return.
NL = 10		# Form feed.
CN = 58		# Colon.

# Limits for elements of HTTP.
METHOD_LENGTH = 10
REQUEST_LENGTH = 8000
HTTP_LENGTH = 16
NAME_LENGTH = 128
VALUE_LENGTH = 1024
PENDING_REQUESTS = 256

SERVER_SLASH_VERSION = 'kipjak-API-server/1.0'
CLIENT_SLASH_VERSION = 'kipjak-API-client/1.0'
SERVER_REST_VERSION = 'kipjak-REST-server/1.0'

LARGE_BODY = 256 * 1024

# Stream the representation of an HTTP request onto the
# byte sequence provided.
def stream_request(encoded_bytes, method=None, request_uri=None, http=None, header=None, body=None):
	method = method or 'POST'
	header = header or {}
	http = http or 'HTTP/1.1'

	if body:
		content_length = len(body)
		request = (f'{method} {request_uri} {http}\r\n'
			f'User-Agent: {CLIENT_SLASH_VERSION}\r\n')

	else:
		content_length = 0
		request = (f'{method} {request_uri} {http}\r\n'
			f'User-Agent: {CLIENT_SLASH_VERSION}\r\n')

	if 'Content-Length' not in header:
		request += f'Content-Length: {content_length}\r\n'

	if 'Content-Type' not in header:
		request += 'Content-Type: application/x-www-form-urlencoded\r\n'

	if header:
		kv = [f'{k}: {v}\r\n' for k, v in header.items()]
		kv = ''.join(kv)
		request += kv
	request += '\r\n'

	encoded_bytes += request.encode('utf-8')
	if body:
		encoded_bytes += body

# Stream the representation of an HTTP response onto the
# byte sequence provided.
def stream_response(encoded_bytes, http=None, status_code=200, reason_phrase=None, header=None, body=None, restful=False):
	http = http or 'HTTP/1.1'
	header = header or {}
	now = world_now()
	date = now.strftime('%a, %d %b %Y %H:%M:%S GMT')

	if reason_phrase is None:
		status_code, reason_phrase = faulted_status(status_code)

	server_version = SERVER_SLASH_VERSION if not restful else SERVER_REST_VERSION
	if body:
		content_length = len(body)
		response = (f'{http} {status_code} {reason_phrase}\r\n'
			f'Date: {date}\r\n'
			f'Server: {server_version}\r\n'
			f'Content-Length: {content_length}\r\n')
	else:
		response = (f'{http} {status_code} {reason_phrase}\r\n'
			f'Date: {date}\r\n'
			f'Server: {server_version}\r\n'
			f'Content-Length: 0\r\n')

	if header:
		kv = [f'{k}: {v}\r\n' for k, v in header.items()]
		kv = ''.join(kv)
		response += kv
	response += '\r\n'

	encoded_bytes += response.encode('utf-8')
	if body:
		encoded_bytes += body

def faulted_status(error_code):
	status_code = error_code or 500
	if status_code < 100:
		status_code = 500
		reason = 'Server Error'
	elif status_code < 200:
		reason = 'Information'
	elif status_code < 300:
		reason = 'OK'
	elif status_code < 400:
		reason = 'Multiple Choices'
	elif status_code < 500:
		reason = 'Client Error'
	elif status_code < 600:
		reason = 'Server Error'
	else:
		status_code = 500
		reason = 'Server Error'
	return status_code, reason

# Inbound requests and outbound responses.
class ApiServerStream(object):
	def __init__(self, transport):
		self.transport = transport

		# Specific to input decoding.
		self.analysis_state = 1
		self.method = bytearray()
		self.request = bytearray()
		self.http = bytearray()

		# Lexical FSMs - REQUEST_LINE.
		def s1(c):
			if c in HTTP_TEXT:
				self.method.append(c)		# GET/POST/...
				return 2
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s2(c):
			if c in HTTP_TEXT:
				n = len(self.method)
				if n < METHOD_LENGTH:
					self.method.append(c)	# GET/POST/... (cont)
					return 2
				return 10
			elif c in HTTP_SPACE:
				return 3
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s3(c):
			if c in HTTP_SPACE:
				return 3
			elif c in HTTP_TEXT:
				self.request.append(c)		# Request URI.
				return 4
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s4(c):
			if c in HTTP_TEXT:
				n = len(self.request)
				if n < REQUEST_LENGTH:
					self.request.append(c)	# Request URI (cont).
					return 4
				return 10
			elif c in HTTP_SPACE:
				return 5
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s5(c):
			if c in HTTP_SPACE:
				return 5
			elif c in HTTP_TEXT:
				self.http.append(c)			# HTTP version.
				return 6
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s6(c):
			if c in HTTP_TEXT:
				n = len(self.http)
				if n < HTTP_LENGTH:
					self.http.append(c)		# HTTP version (cont).
					return 6
				return 10
			elif c in HTTP_SPACE:
				return 9
			elif c == CR:
				return 8
			elif c == NL:
				return REQUEST_LINE			# Shortcut request line.
			return 10

		def s8(c):
			if c == NL:
				return REQUEST_LINE			# Request line.
			return 10

		def s9(c):
			if c in HTTP_SPACE:
				return 9
			elif c == CR:
				return 8
			elif c == NL:
				return REQUEST_LINE			# Shortcut request line.
			return 10

		# Failed the proper layout, discard up to next line termination.
		def s10(c):
			if c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s11(c):
			if c == NL:
				return DISCARD_LINE
			return 10

		# HEADER_LINE.
		def s20(c):
			if c in HTTP_TEXT:
				self.name.append(c)		# Header name.
				return 21
			elif c in HTTP_SPACE:
				return 40
			elif c == CR:
				return 41
			elif c == NL:
				return EMPTY_LINE
			return 10

		def s21(c):
			if c == CN:					# Colon. MUST check this first.
				return 22
			elif c in HTTP_TEXT:
				n = len(self.name)
				if n < NAME_LENGTH:
					self.name.append(c)	# Header name (cont).
					return 21
				return 10
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s22(c):
			if c in HTTP_SPACE:		# Between colon and value.
				return 23
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s23(c):
			if c in HTTP_PRINTING:
				self.value.append(c)	# Header value.
				return 24
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s24(c):
			if c in HTTP_PRINTING:
				n = len(self.value)
				if n < VALUE_LENGTH:
					self.value.append(c)	# Header value (cont).
					return 24
				return 10
			elif c in HTTP_SPACE:
				return 27
			elif c == CR:
				return 26
			elif c == NL:
				return HEADER_LINE
			return 10

		def s26(c):
			if c == NL:
				return HEADER_LINE
			return 10

		def s27(c):
			if c in HTTP_SPACE:
				return 27
			elif c == CR:
				return 26
			elif c == NL:
				return HEADER_LINE
			return 10

		# EMPTY_LINE
		def s40(c):
			if c in HTTP_SPACE:
				return 40
			elif c == CR:
				return 41
			elif c == NL:
				return EMPTY_LINE
			return 10

		def s41(c):
			if c == NL:
				return EMPTY_LINE
			return 10

		# BODY_BLOCK
		def s30(c):
			self.body.append(c)
			self.length -= 1
			if self.length < 1:
				return BODY_BLOCK
			return 30

		self.shift = {
			1: s1, 2: s2, 3: s3, 4: s4, 5: s5,
			6: s6, 8: s8, 9: s9, 10: s10, 11: s11,

			20: s20, 21: s21, 22: s22, 23: s23,
			24: s24, 26: s26, 27: s27,

			40: s40, 41: s41,

			30: s30,
		}

	# Message being sent from HTTP server.
	# Push the optimal representation of the next message onto the byte stream.
	def message_to_block(self, mtr):
		m, t, r = mtr
		transport = self.transport
		encoded_bytes = transport.encoded_bytes
		codec = transport.codec
		search_subs = self.transport.parent.search_subs
		restful = search_subs is not None

		content_json = {'Content-Type': 'application/json'}
		content_text = {'Content-Type': 'plain/text'}
		if isinstance(m, Faulted):
			status_code, reason_phrase = faulted_status(m.error_code)
			e = codec.encode(m, Any())
			body = e.encode('utf-8')
			stream_response(encoded_bytes,
				status_code=status_code, reason_phrase=reason_phrase,
				header=content_json,
				body=body, restful=restful)
		elif not isinstance(m, HttpResponse):
			e = codec.encode(m, Any())
			body = e.encode('utf-8')
			stream_response(encoded_bytes,
				status_code=200, reason_phrase='OK',
				header=content_json,
				body=body, restful=restful)
		elif m.body is not None:
			# Special case handling; allow sending of text and
			# specific content type and payload.
			if isinstance(m.body, bytearray):
				if 'Content-Type' not in m.header:
					body = f'Custom response but "Content-Type" not specified'.encode('utf-8')
					stream_response(encoded_bytes, status_code=500, reason_phrase='Server Error',
						header=content_text, body=body, restful=restful)
					return
				stream_response(encoded_bytes,
					http=m.http, status_code=m.status_code, reason_phrase=m.reason_phrase,
					header=m.header, body=m.body, restful=restful)
				return
			elif isinstance(m.body, str):
				m.header['Content-Type'] = 'plain/text'
				body = m.body.encode('utf-8')
				stream_response(encoded_bytes,
					http=m.http, status_code=m.status_code, reason_phrase=m.reason_phrase,
					header=m.header, body=body, restful=restful)
				return

			# Standard processing. The body is an instance of Any,
			# i.e. either a class instance or a constructed type.
			m.header['Content-Type'] = 'application/json'
			e = codec.encode(m.body, Any())
			body = e.encode('utf-8')
			stream_response(encoded_bytes,
				http=m.http, status_code=m.status_code, reason_phrase=m.reason_phrase,
				header=m.header, body=body, restful=restful)
		else:
			stream_response(encoded_bytes,
				http=m.http, status_code=m.status_code, reason_phrase=m.reason_phrase,
				header=m.header, restful=restful)

	# Complete zero or more messages, using the given block.
	def recover_message(self, received, sockets):
		for c in received:
			next = self.shift[self.analysis_state](c)
			if next > 0:
				self.analysis_state = next
				continue

			# Poor man's reduce-goto.
			if next == REQUEST_LINE:
				self.analysis_state = 20
				self.name = bytearray()
				self.value = bytearray()
				self.header = {}
				self.body = None
				self.length = 0
				continue
			elif next == HEADER_LINE:
				name = self.name.decode('ascii')
				value = self.value.decode('ascii')
				self.header[name] = value
				self.analysis_state = 20
				self.name = bytearray()
				self.value = bytearray()
				continue
			elif next == EMPTY_LINE:
				s = self.header.get('Content-Length', None)
				if s:
					self.length = int(s)
					if self.length > LARGE_BODY:
						# Discard everything. Should close. Likely
						# to land in nowheres-ville and fail anyway.
						self.analysis_state = 1
						self.method = bytearray()
						self.request = bytearray()
						self.http = bytearray()
						continue
					self.body = bytearray()
					if self.length > 0:
						self.analysis_state = 30
						continue
			elif next == BODY_BLOCK:
				pass
			else:
				# Discard line,
				self.analysis_state = 1
				self.method = bytearray()
				self.request = bytearray()
				self.http = bytearray()
				continue

			body, to_address, return_address = self.recover_hyper_text()

			# Completed frame.
			yield body, to_address, return_address

			# Restart.
			self.analysis_state = 1
			self.method = bytearray()
			self.request = bytearray()
			self.http = bytearray()

	# An HTTP request has been received by an HTTP server.
	# Recover the best-possible message and forward to receiver.
	def recover_hyper_text(self):
		transport = self.transport
		codec = transport.codec
		return_address = transport.return_proxy
		to_address = transport.local_termination

		# Parsed elements of the HTTP request
		method = self.method.decode('ascii')
		request = self.request.decode('ascii')
		http = self.http.decode('ascii')
		header = self.header
		content_type = self.header.get('Content-Type', None)
		body = self.body

		# Handling details.
		named_type = self.transport.parent.named_type
		search_subs = self.transport.parent.search_subs
		default_to_request = self.transport.parent.request.default_to_request

		try:	# Raise exception for immediate error response.

			if search_subs:
				search, subs = search_subs
				match = search.fullmatch(request)
				if match:
					form_entry = {s: match[s] for s in subs}
					message = FormRequest(method, request, form_entry, http, header, body)
					return message, to_address, return_address
				raise ValueError(f'request URI does not match form')

			elif not request.startswith('/') or len(request) < 2:
				raise ValueError(f'unexpected request URI "{request}"')

			slash = request.rfind('/')			# Test above guarantees a slash.
			name = request[slash + 1:]			# Only want the trailing name.
			question = name.find('?')
			if question != -1:
				query = name[question+1:]
				name = name[:question]
			else:
				query = ''

			tom = named_type.get(name, None)
			if tom is None:
				if not default_to_request:
					raise ValueError(f'unknown and defaulting to HTTP requests disabled')
			elif question != -1:
				schema = tom.__art__.schema
				message = tom()
				a = query.split('&')
				try:
					c = CodecJson()
					for kv in a:
						if not kv:
							continue
						equals = kv.find('=')
						if equals < 0:
							raise ValueError(f'unexpected field "{kv}" query')
						k = kv[:equals]
						v = kv[equals + 1:]
						q = unquote_plus(v)

						t = schema.get(k, None)
						if t is None:
							raise ValueError(f'unknown key "{k}" in "{name}" query')
						d = decode_argument(c, q, t)
						setattr(message, k, d)
				except CodecError as e:
					s = str(e)
					raise ValueError(f'no conversion for "{k}" ({s})')
				return message, to_address, return_address
			elif content_type == 'application/x-www-form-urlencoded':
				schema = tom.__art__.schema
				message = tom()
				if body:
					# Breakout the flat form into a key-value dict.
					d = body.decode('utf-8')
					a = d.split('&')

					try:
						c = CodecJson()
						for kv in a:
							equals = kv.find('=')
							if equals < 0:
								raise ValueError(f'unexpected field "{kv}" x-www-form')
							k = kv[:equals]
							v = kv[equals + 1:]
							q = unquote_plus(v)

							t = schema.get(k, None)
							if t is None:
								raise ValueError(f'unknown key "{k}" in "{name}" x-www-form')

							d = decode_argument(c, q, t)
							setattr(message, k, d)
					except CodecError as e:
						s = str(e)
						raise ValueError(f'no conversion for "{k}" ({s})')

				return message, to_address, return_address

			elif content_type == 'application/json':
				# Unlikely that 3rd party JSON will be in kipjak layout.
				# This is effectively for kipjak clients.
				if method != 'POST':
					raise ValueError(f'unexpected HTTP method "{method}"')
				if body:
					json = body.decode('utf-8')
					try:
						message = codec.decode(json, UserDefined(tom))
					except (CodecError, OverflowError, ValueError) as e:
						s = str(e)
						raise ValueError(s)
				else:
					message = tom()
				return message, to_address, return_address
			else:
				# Matched in the API but not a content that
				# can be processed
				raise ValueError(f'unexpected content type "{content_type}" for HTTP request "{name}"')

		except (ValueError, ConversionError) as e:
			s = str(e)
			s = f'cannot recover request "{request}" ({s})'
			message = HttpResponse(status_code=400, reason_phrase='Bad Request', body=s)
			return message, return_address, to_address

		# Default is to pass on any request that parsed correctly
		# but did not meet semantic criteria.
		message = HttpRequest(method=method, request_uri=request, http=http,
			header=self.header,
			body=self.body)
		return message, to_address, return_address

#
#
class INITIAL: pass
class READY: pass

class ApiClientSession(Point, StateMachine):
	def __init__(self, proxy_address=None, **kv):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.proxy_address = proxy_address
		self.pending = deque()

def ApiClientSession_INITIAL_Start(self, message):
	return READY

def ApiClientSession_READY_Unknown(self, message):
	# A message passing through, either a response from the
	# remote end or another request from a local object.
	if self.return_address[-1] == self.proxy_address[-1]:	# Response from remote.
		if self.pending:									# Yes there is a matching request.
			m, t, r = self.pending.popleft()
			c = cast_to(message, self.received_type)
			self.send(c, r)									# Send response to original client.
			if self.pending:								# A waiting request?
				m, t, r = self.pending[0]
				c = cast_to(m, t)
				self.send(m, self.proxy_address)
		else:
			#t = tof(message)
			self.warning(f'message "tof" from HTTP server has no matching request')
	elif len(self.pending) > PENDING_REQUESTS:
		self.reply(HttpResponse(status_code=400, reason_phrase='Client Error', body='Request queue overflow'))
	else:
		mtr = (message, self.received_type, self.return_address)				# Request from local client.
		self.pending.append(mtr)							# Remember.
		if len(self.pending) == 1:						# Nothing pending.
			self.send(message, self.proxy_address)
	return READY

def ApiClientSession_READY_Stop(self, message):
	self.complete(Aborted())

API_CLIENT_SESSION_DISPATCH = {
	INITIAL: (
		(Start,), ()
	),
	READY: (
		(Unknown, Stop), ()
	),
}

bind(ApiClientSession, API_CLIENT_SESSION_DISPATCH, thread='api-client-session')

#
#
class ApiClientStream(object):
	def __init__(self, transport):
		self.transport = transport

		# Specific to input decoding.
		self.analysis_state = 1

		# Response.
		self.http = bytearray()
		self.code = bytearray()
		self.reason = bytearray()
		# Header.
		self.name = bytearray()
		self.value = bytearray()
		self.header = {}
		# Body.
		self.body = None

		# RESPONSE_LINE.
		def s1(c):
			if c in HTTP_TEXT:
				self.http.append(c)		# HTTP/...
				return 2
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s2(c):
			if c in HTTP_TEXT:
				n = len(self.http)
				if n < METHOD_LENGTH:
					self.http.append(c)	# HTTP/...
					return 2
				return 10
			elif c in HTTP_SPACE:
				return 3
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s3(c):
			if c in HTTP_SPACE:
				return 3
			elif c in HTTP_DIGIT:
				self.code.append(c)		# Response code.
				return 4
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s4(c):
			if c in HTTP_DIGIT:
				n = len(self.code)
				if n < 8:
					self.code.append(c)	# Response code.
					return 4
				return 10
			elif c in HTTP_SPACE:
				return 5
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s5(c):
			if c in HTTP_SPACE:
				return 5
			elif c in HTTP_PRINTING:
				self.reason.append(c)		# Reason phrase.
				return 7
			elif c == CR:
				return 8
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s7(c):
			if c in HTTP_PRINTING:
				n = len(self.reason)
				if n < HTTP_LENGTH:
					self.reason.append(c)	# Reason phrase.
					return 7
				return 10
			elif c == CR:
				return 8
			elif c == NL:
				return RESPONSE_LINE			# Shortcut response line.
			return 10

		def s8(c):
			if c == NL:
				return RESPONSE_LINE			# Response line.
			return 10

		# Failed the proper layout, discard up to next line termination.
		def s10(c):
			if c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s11(c):
			if c == NL:
				return DISCARD_LINE
			return 10

		# HEADER_LINE.
		def s20(c):
			if c in HTTP_TEXT:
				self.name.append(c)		# Header name.
				return 21
			elif c in HTTP_SPACE:
				return 40
			elif c == CR:
				return 41
			elif c == NL:
				return EMPTY_LINE
			return 10

		def s21(c):
			if c == CN:		# Colon. Check this first.
				return 22
			elif c in HTTP_TEXT:
				n = len(self.name)
				if n < NAME_LENGTH:
					self.name.append(c)	# Header name.
					return 21
				return 10
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s22(c):
			if c in HTTP_SPACE:		# Between colon and value.
				return 23
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s23(c):
			if c in HTTP_PRINTING:
				self.value.append(c)	# Header value.
				return 24
			elif c == CR:
				return 11
			elif c == NL:
				return DISCARD_LINE
			return 10

		def s24(c):
			if c in HTTP_PRINTING:
				n = len(self.value)
				if n < VALUE_LENGTH:
					self.value.append(c)	# Header value.
					return 24
				return 10
			elif c in HTTP_SPACE:
				return 27
			elif c == CR:
				return 26
			elif c == NL:
				return HEADER_LINE
			return 10

		def s26(c):
			if c == NL:
				return HEADER_LINE
			return 10

		def s27(c):
			if c in HTTP_SPACE:
				return 27
			elif c == CR:
				return 26
			elif c == NL:
				return HEADER_LINE
			return 10

		# EMPTY_LINE
		def s40(c):
			if c in HTTP_SPACE:
				return 40
			elif c == CR:
				return 41
			elif c == NL:
				return EMPTY_LINE
			return 10

		def s41(c):
			if c == NL:
				return EMPTY_LINE
			return 10

		# BODY_BLOCK
		def s30(c):
			self.body.append(c)
			self.length -= 1
			if self.length < 1:
				return BODY_BLOCK
			return 30

		self.shift = {
			1: s1, 2: s2, 3: s3, 4: s4, 5: s5,
			7: s7, 8: s8,

			10: s10, 11: s11,

			20: s20, 21: s21, 22: s22, 23: s23,
			24: s24, 26: s26, 27: s27,

			40: s40, 41: s41,

			30: s30,
		}

	# Message being sent from HTTP client.
	# Push the optimal representation of the next message onto the byte stream.
	def	message_to_block(self, mtr):
		m, t, r = mtr
		transport = self.transport
		codec = transport.codec
		encoded_bytes = transport.encoded_bytes
		http_client = transport.parent.request.http_client
		application_json = transport.parent.request.application_json
		tom = type(m)
		art = getattr(tom, '__art__', None)
		if art is None:
			raise ValueError('HTTP client output limited to class instances (including HttpResponse)')

		# Client is sending an explicit HTTP request.
		# Add the barest automation, e.g. prepend path, content type and length.
		if isinstance(m, HttpRequest):
			request_uri = f'{http_client}{m.request_uri}'
			stream_request(encoded_bytes,
				method=m.method, request_uri=request_uri, http=m.http,
				header=m.header, body=m.body)
			return

		if application_json:
			content_type = 'application/json'
			e = codec.encode(m, UserDefined(tom))
			body = e.encode('utf-8')
		else:
			content_type = 'application/x-www-form-urlencoded'
			schema = tom.__art__.schema
			ks = []
			try:
				c = CodecNoop()
				for k, t in schema.items():
					# Form representation of null is to omit.
					d = getattr(m, k, None)
					if d is None:
						continue
					s = encode_argument(c, d, t)
					q = quote_plus(s)
					ks.append(f'{k}={q}')
			except (CodecError, ValueError) as e:
				raise ValueError(f'cannot stream "{art.path}" as x-www-form, {e}')

			if ks:
				form = '&'.join(ks)
				body = form.encode('utf-8')
			else:
				body = None

		request_uri = f'{http_client}{art.name}'
		stream_request(encoded_bytes,
			method='POST', request_uri=request_uri,
			header={'Content-Type': content_type},
			body=body)

	# Complete zero or more messages, consuming the given block.
	def recover_message(self, received, sockets):
		for c in received:
			next = self.shift[self.analysis_state](c)
			if next > 0:
				self.analysis_state = next
				continue

			# Poor man's reduce-goto.
			if next == RESPONSE_LINE:
				self.analysis_state = 20
				self.name = bytearray()
				self.value = bytearray()
				self.header = {}
				self.body = None
				self.length = 0
				continue
			elif next == HEADER_LINE:
				name = self.name.decode('ascii')
				value = self.value.decode('ascii')
				self.header[name] = value
				self.analysis_state = 20
				self.name = bytearray()
				self.value = bytearray()
				continue
			elif next == EMPTY_LINE:
				s = self.header.get('Content-Length', None)
				if s:
					self.length = int(s)
					self.body = bytearray()
					if self.length > 0:
						self.analysis_state = 30
						continue
			elif next == BODY_BLOCK:
				pass
			else:
				# Discard line,
				self.analysis_state = 1
				self.http = bytearray()
				self.code = bytearray()
				self.reason = bytearray()
				continue

			body, to_address, return_address = self.recover_hyper_text()

			# Completed frame.
			yield body, to_address, return_address

			# Restart.
			self.analysis_state = 1
			self.http = bytearray()
			self.code = bytearray()
			self.reason = bytearray()

	# HTTP response has been received by this client.
	# Generate the optimal message and forward to receiver.
	def recover_hyper_text(self):
		transport = self.transport
		codec = transport.codec

		return_address = transport.return_proxy
		to_address = transport.local_termination

		application_json = transport.parent.request.application_json

		http = self.http.decode('ascii')
		code = self.code.decode('ascii')
		code = int(code)
		reason = self.reason.decode('ascii')
		content_type = self.header.get('Content-Type', None)
		body = self.body

		if not application_json:
			# Compile elements into local message and send.
			message = HttpResponse(http=http, status_code=code, reason_phrase=reason,
				header=self.header,
				body=self.body)
			return message, to_address, return_address

		# Solely for when client is connected to kipjak server.
		# Preconditions for json decode.
		client_error = None
		if http != 'HTTP/1.1':
			client_error = f'unexpected HTTP version "{http}"'
		elif content_type != 'application/json':
			message = HttpResponse(http=http, status_code=code, reason_phrase=reason,
				header=self.header,
				body=self.body)
			return message, to_address, return_address
		elif not body or not isinstance(body, bytearray):
			client_error = f'no JSON body'
		else:
			v = None
			try:
				body = body.decode('utf-8')
				message = codec.decode(body, Any())
			except (CodecError, OverflowError, ValueError) as e:
				s = str(e)
				client_error = f'unexpected body "{s}"'

		if client_error:
			message = HttpResponse(http=http, status_code=500, reason_phrase='Server Error', body=client_error)
			return message, to_address, return_address

		return message, to_address, return_address



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Capture the standard information from the command line.

An object that defines the values that may be gathered from sys.argv during
standard processing of the command line.
"""
__docformat__ = 'restructuredtext'

from enum import Enum
from .ip_networking import *
from .virtual_memory import *
from .message_memory import *
from .virtual_runtime import *

__all__ = [
	'ScopeOfDirectory',
	'ProcessOrigin',
	'CommandLine',
	'CL',
]

# Register service or interest in service.
class ScopeOfDirectory(Enum):
	"""
	Enumeration of the scopes within the **kipjak** directory.

	* LAN - local area network
	* HOST - networking limited to 127.0.0.0
	* GROUP - a composite process
	* PROCESS - a single process
	* LIBRARY - dynamically loaded process
	"""
	WAN=1
	LAN=2
	HOST=3
	GROUP=4
	PROCESS=5
	LIBRARY=6

class ProcessOrigin(Enum):
	"""
	Enumeration of context in the chain of processes.

	* SHELL - running from the command-line
	* RUN - run an entry in a composite process
	* RUN_CHILD - run as a subprocess of a composite entry
	* START - background run, entry in a composite process
	* START_CHILD - background run, subprocess of a composite entry
	"""
	SHELL=0
	RUN=1
	RUN_CHILD=2
	START=3
	START_CHILD=4

#
class CommandLine(object):
	"""
	Standard instructions to a new process.

	Communicate the context for the new process, e.g. daemon, home/role,
	settings and parent/child communications.

	:param origin: position in the process/sub-process chain
	:param child_process: this process is a standard child of a standard parent
	:param full_output: enable full parent-child process integration
	:param debug_level: select the level of logs to display
	:param home_path: location of the process group
	:param role_name: role within a process group
	:param resource_path: override the location of per-executable, read-only resources
	:param model_path: override the location of per-instance read-write materials
	:param help: enable output of help page
	:param create_role: save the specified settings
	:param update_role: add/override the specified settings
	:param dump_role: enable output of current settings
	:param edit_role: enable visual editing of the current settings
	:param factory_reset: discard stored settings
	:param delete_role: remove persisted settings
	:param role_file: use the settings in the specified file
	:param dump_types: enable output of type table
	:param output_file: place any output in the specified file
	:param directory_scope: scope of this process
	:param connect_to_directory: IP and port of parent directory
	:param encrypted_process: enable encryption of connection to parent directory
	:param accept_directories_at: IP and port where child directories are accepted
	"""
	def __init__(self,
			origin: ProcessOrigin=ProcessOrigin.SHELL,
			child_process: bool=False,
			full_output: bool=False,
			debug_level=None,
			home_path: str=None, role_name: str=None,
			resource_path: str=None, model_path: str=None,
			help: bool=False,
			create_role: bool=False, update_role: bool=False,
			factory_reset: bool=False,
			dump_role: bool=False,
			edit_role: bool=False,
			delete_role: bool=False,
			role_file: str=None,
			dump_types: bool=False,
			output_file: str=None,
			keep_logs: bool=False,
			directory_scope: ScopeOfDirectory=None,
			connect_to_directory: HostPort=None,
			accept_directories_at: HostPort=None,
			encrypted_process: bool=False):
		self.origin = origin
		self.child_process = child_process
		self.full_output = full_output
		self.debug_level = debug_level
		self.home_path = home_path
		self.role_name = role_name
		self.resource_path = resource_path
		self.model_path = model_path
		self.help = help
		self.create_role = create_role
		self.update_role = update_role
		self.factory_reset = factory_reset
		self.dump_role = dump_role
		self.edit_role = edit_role
		self.delete_role = delete_role
		self.role_file = role_file
		self.dump_types = dump_types
		self.output_file = output_file
		self.keep_logs = keep_logs
		self.directory_scope = directory_scope or ScopeOfDirectory.PROCESS
		self.connect_to_directory = connect_to_directory or HostPort()
		self.accept_directories_at = accept_directories_at or HostPort()
		self.encrypted_process = encrypted_process

bind_message(CommandLine,
	debug_level=Enumeration(USER_LOG),
)

CL = CommandLine()



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Management of the async runtime.

Ensure that the support for async operation is in place when the process
needs it. Ensure that support is cleared out during process termination.
"""
__docformat__ = 'restructuredtext'

import os
import tempfile
from .virtual_memory import *
from .virtual_codec import *
from .json_codec import *
from .noop_codec import *
from .file_object import *
from .point_runtime import *
from .home_role import *
from .process_object import *


def edit_role(self, home, editor=None):
	testing = editor == 'test-passthru-editor'
	editor = editor or os.getenv('LC_EDITOR') or 'vi'

	try:
		fd, name = tempfile.mkstemp()
		os.close(fd)

		# Prepare materials for editor.
		temporary = File(name, MapOf(Unicode(), Any()), decorate_names=False)
		temporary.store(home.settings())

		# Setup detection of change.
		modified = os.stat(name).st_mtime

		# Run the editor.
		a = self.create(Utility, editor, name)
		self.assign(a, editor)
		m, i = self.select(Returned, Faulted, Stop)
		e = self.debrief()
		if isinstance(m, Faulted) and not testing:
			return m
		message = m.message
		if isinstance(message, Faulted) and not testing:
			return message

		# Was the file modified?
		if os.stat(name).st_mtime == modified:
			return Faulted(f'settings not modified')

		# Validate contents and update the runtime.
		r = temporary.recover()
		home.settings.update(r)
	except (CodecError, OSError) as e:
		return Faulted(f'cannot update settings ({e})')
	finally:
		os.remove(name)
	return None

HR.edit_role = edit_role



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Directory at the GROUP scope.

Starts and monitors a collection of processes, according to the materials
it finds in a folder (i.e. home_path) and optional search criteria.

An empty list of patterns implies "all roles". A non-empty list of
patterns that fails to match anything is considered an error.

Termination is by intervention (i.e. control-c) or termination of the
child processes. There are several variations on this general theme;

* termination of all processes, retries optional (the default),
* termination of a specified process, retries optional (see main_role).
"""
__docformat__ = 'restructuredtext'

import re
import kipjak as kj

#
#
DEFAULT_GROUP = 'group'

class INITIAL: pass
class ENQUIRING: pass
class RUNNING: pass
class RETURNING: pass
class GROUP_RETURNING: pass

dict_type = kj.def_type(dict[str,kj.Any])

class Group(kj.Threaded, kj.StateMachine):
	def __init__(self, *search,
			directory_at_host: kj.HostPort=None, directory_at_lan: kj.HostPort=None,
			encrypted_directory: bool=None,
			retry: kj.RetryIntervals=None, main_role: str=None) -> kj.Any:
		kj.Threaded.__init__(self)
		kj.StateMachine.__init__(self, INITIAL)
		self.search = search			# List or re's.
		self.directory_at_host = directory_at_host
		self.directory_at_lan = directory_at_lan
		self.encrypted_directory = encrypted_directory
		self.retry = retry
		self.main_role = main_role

		self.home_path = kj.CL.home_path or kj.DEFAULT_HOME
		self.role_name = kj.CL.role_name or DEFAULT_GROUP

		self.machine = []				# Compiled re's.
		self.home = {}					# Roles matched by patterns.
		self.ephemeral = None			# The connect_to_directory value for child processes.
		self.interval = {}				# Interval iterators for each role.
		self.group_returned = {}		# Values returned by self-terminating processes

def Group_INITIAL_Start(self, message):
	if self.search:
		s = ', '.join(self.search)
		self.trace(f'Search "{s}"')

	if self.directory_at_host:
		connect_to_directory = self.directory_at_host
	elif self.directory_at_lan:
		connect_to_directory = self.directory_at_lan
	else:
		connect_to_directory = None

	accept_directories_at = kj.HostPort('127.0.0.1', 0)

	self.directory = self.create(kj.ObjectDirectory, directory_scope=kj.ScopeOfDirectory.GROUP,
		connect_to_directory=connect_to_directory,
		accept_directories_at=accept_directories_at,
		encrypted=self.encrypted_directory)
	self.assign(self.directory, 0)

	self.send(kj.Enquiry(), self.directory)
	return ENQUIRING

def Group_ENQUIRING_HostPort(self, message):
	self.ephemeral = message					# This is accept_directories_at.

	# Load all the roles.
	home = kj.open_home(self.home_path)
	if home is None:
		self.complete(kj.Faulted(f'cannot open path "{self.home_path}"'))

	# Compile all the patterns.
	for s in self.search:
		try:
			m = re.compile(s)
		except re.error as e:
			self.complete(kj.Faulted(f'cannot compile search "{s}"', str(e)))
		self.machine.append(m)

	# Scan for roles matching a pattern.
	if self.machine:
		def match(name):
			for m in self.machine:
				b = m.match(name)
				if b:
					return True
			return False

		home = {k: v for k, v in home.items() if match(k)}
		if not home:
			s = ', '.join(self.search)
			self.complete(kj.Faulted(f'No roles matching "{s}"'))

	elif not home:
		self.complete(kj.Faulted(f'No roles at location "{self.home_path}"'))

	# Start the roles in this non-empty list.
	encrypted_process = self.encrypted_directory == True
	for k, v in home.items():
		a = self.create(kj.ProcessObject, v,
			home_path=self.home_path, role_name=k, top_role=True,
			directory_scope=kj.ScopeOfDirectory.PROCESS,
			connect_to_directory=self.ephemeral,
			encrypted_process=encrypted_process)
		self.assign(a, k)

	# Remember for restarts.
	self.home = home
	return RUNNING

def Group_ENQUIRING_Faulted(self, message):
	self.complete(message)

def Group_ENQUIRING_Stop(self, message):
	self.complete(kj.Aborted())

def Group_RUNNING_Returned(self, message):
	d = self.debrief()
	if isinstance(d, kj.OnReturned):			# Restart callbacks.
		d(self, message)
		return RUNNING

	if d == self.main_role:						# Declared "main" - no retries.
		if not self.working():					# Includes ProcessObjects and restart callbacks.
			self.complete(message.message)

		self.abort(message.message)
		return RETURNING

	#
	self.group_returned[d] = message.message

	if self.retry is None:						# Not configured for restarts.
		if not self.working():					# As above.
			message = kj.cast_to(self.group_returned, dict_type)
			self.complete(message)

		if self.main_role is None:
			self.abort()
			return GROUP_RETURNING

		return RUNNING

	i = self.interval.get(d, None)
	if i is None:
		i = kj.smart_intervals(self.retry)
		self.interval[d] = i

	try:
		seconds = next(i)
		self.trace(f'Restart "{d}" ({seconds} seconds)')
	except StopIteration:
		if not self.working():					# As above.
			message = kj.cast_to(self.group_returned, dict_type)
			self.complete(message)

		if self.main_role is None:
			self.abort()
			return GROUP_RETURNING

		return RUNNING

	def restart(self, value, args):
		a = self.create(kj.ProcessObject, self.home[args.role],
			home_path=self.home_path, role_name=args.role, top_role=True,
			directory_scope=kj.ScopeOfDirectory.PROCESS, connect_to_directory=self.ephemeral)
		self.assign(a, args.role)

	# Run a no-op with the desired timeout.
	a = self.create(kj.Delay, seconds=seconds)
	self.on_return(a, restart, role=d)
	return RUNNING

def Group_RUNNING_Faulted(self, message):
	if not self.working():
		self.complete(message)

	self.abort(message.message)
	return RETURNING

def Group_RUNNING_Stop(self, message):
	if not self.working():
		self.complete(kj.Aborted())

	self.abort(kj.Aborted())
	return RETURNING

def Group_RETURNING_Returned(self, message):
	d = self.debrief()
	if isinstance(d, kj.OnReturned):		# Restart callbacks.
		pass

	if not self.working():					# Includes ProcessObjects and restart callbacks.
		self.complete(self.aborted_message)

	return RETURNING

def Group_GROUP_RETURNING_Returned(self, message):
	d = self.debrief()
	if isinstance(d, kj.OnReturned):		# Restart callbacks.
		pass
	else:
		self.group_returned[d] = message.message

	if not self.working():					# Includes ProcessObjects and restart callbacks.
		message = kj.cast_to(self.group_returned, dict_type)
		self.complete(message)

	return GROUP_RETURNING


GROUP_DISPATCH = {
	INITIAL: (
		(kj.Start,),
		()
	),
	ENQUIRING: (
		(kj.HostPort, kj.Faulted, kj.Stop),
		()
	),
	RUNNING: (
		(kj.Returned, kj.Faulted, kj.Stop),
		()
	),
	RETURNING: (
		(kj.Returned,),
		()
	),
	GROUP_RETURNING: (
		(kj.Returned,),
		()
	),
}

kj.bind(Group, GROUP_DISPATCH)

def main():
	# See send(kj.Enquiry()) and Group_ENQUIRING_HostPort for how
	# this process acquires the right listening configuration.
	# Unless there is an explicit argument this will open a listen
	# port at 127.0.0.1:0 (i.e. ephemeral). If the directory
	# is presented with pub-subs for higher levels and no connect
	# address has been specified, it will auto-connect to
	# DIRECTORY_AT_HOST (e.g. 127.0.0.1:DIRECTORY_PORT)
	kj.create(Group)

if __name__ == '__main__':
	main()
